#include "constraint.h"
#include "fragment.h"


/* Utility functions: */
Fragment*
makeIntersection (const Fragment& fragA, 
		  guint32 bytesUsedA, 
		  const Fragment& fragB,
		  guint32 bytesUsedB,
		  guint32 bytesToUse)
{
	g_assert (bytesUsedA < fragA.getLength());
	g_assert (bytesUsedA+bytesToUse <= fragA.getLength());
	g_assert (bytesUsedB < fragB.getLength());
	g_assert (bytesUsedB+bytesToUse <= fragB.getLength());

	const FragmentFixed *fixedA = dynamic_cast<const FragmentFixed *>(&fragA);
	const FragmentFixed *fixedB = dynamic_cast<const FragmentFixed *>(&fragB);

	if (fixedA) {
		if (fixedB) {
			/* Both are fixed; if they have the same content over the given run, then clone the relevant bytes into a new "fixed", otherwise no intersection exists: */
			for (int i=0;i<bytesToUse;i++) {
				if (fixedA->getByte(i) != fixedB->getByte(i)) {
					return NULL;
				}
		
			}

			/* They have the same content over the range; use A: */
			return fixedA->cloneRange (bytesUsedA,
						   bytesToUse);
		} else {
			/* A is "fixed", B is "any", clone the relevant bytes of A into a new "fixed": */
			return fixedA->cloneRange (bytesUsedA,
						   bytesToUse);
		}
	} else {
		if (fixedB) {
			/* A is "any", B is "fixed", clone the relevant bytes of B into a new "fixed": */
			return fixedB->cloneRange (bytesUsedB,
						   bytesToUse);
		} else {
			/* Both are "any"; make an "any" of the correct length: */
			return new FragmentAny (bytesToUse);
		}
	}
}

void
reportImpossible (const FragmentList& inputA,
		  const FragmentList& inputB)
{
	//	g_message ("got impossible combination");
}


/* Returns NULL if not possible to combine: */
FragmentList*
combineFragmentLists (const FragmentList& inputA,
		      const FragmentList& inputB)
{
#if 0
	g_message ("combineFragmentLists");
#endif

	FragmentList* newList = new FragmentList;

	guint32 bytesUsedA=0;
	guint32 bytesUsedB=0;
	FragmentList::const_iterator iterA = inputA.begin();
	FragmentList::const_iterator iterB = inputB.begin();

	while ((iterA!=inputA.end()) || (iterB!=inputB.end())) {
		if ((iterA!=inputA.end()) && (iterB!=inputB.end())) {
			/* Then we still have fragments from both input to process: */
			/* which frag finishes next? */

			g_assert (bytesUsedA < (*iterA)->getLength());
			g_assert (bytesUsedB < (*iterB)->getLength());

			if (((*iterA)->getLength()-bytesUsedA)<((*iterB)->getLength()-bytesUsedB)) {
				/* A finishes first: */
				guint32 bytesToUse = ((*iterA)->getLength()-bytesUsedA);
				Fragment *intersection = makeIntersection (**iterA, 
									   bytesUsedA, 
									   **iterB, 
									   bytesUsedB,
									   bytesToUse);
				if (intersection) {
					newList->push_back (RefCountPtr<Fragment> (intersection));
				} else {
					/* Impossible: */
					reportImpossible (inputA, inputB);
					return NULL;
				}
				iterA++;
				bytesUsedA = 0;
				bytesUsedB += bytesToUse;

			} else {
				/* B finishes first (or simultaneous): */
				guint32 bytesToUse = ((*iterB)->getLength()-bytesUsedB);
				Fragment *intersection = makeIntersection (**iterA, 
									   bytesUsedA, 
									   **iterB, 
									   bytesUsedB,
									   bytesToUse);
				if (intersection) {
					newList->push_back (intersection);
				} else {
					/* Impossible: */
					reportImpossible (inputA, inputB);
					return NULL;
				}
				iterB++;
				bytesUsedA += bytesToUse;
				bytesUsedB = 0;
				
				/* Handle the case where they finish simultaneously: */
				if (bytesUsedA == (*iterA)->getLength()) {
					iterA++;
					bytesUsedA=0;
				}
			}
		} else {
			if (iterA!=inputA.end()) {
				/* A has remaining fragments (and B doesn't): */
				if (bytesUsedA>0) {
					guint32 bytesToUse = ((*iterA)->getLength()-bytesUsedA);
					
					Fragment *clone = (*iterA)->cloneRange (bytesUsedA,
										bytesToUse);
					newList->push_back (clone);
					bytesUsedA = 0;
				} else {
					newList->push_back (*iterA);
				}
				iterA++;
			} else {
				/* B has remaining fragments (and A doesn't): */
				if (bytesUsedB>0) {
					guint32 bytesToUse = ((*iterB)->getLength()-bytesUsedB);
					
					Fragment *clone = (*iterB)->cloneRange (bytesUsedB,
										bytesToUse);
					newList->push_back (clone);
					bytesUsedB = 0;
				} else {
					newList->push_back (*iterB);
				}
				iterB++;
			}
		}		
	}

	return newList;
}

Constraint::Result_t
intersectResults (Constraint::Result_t result0, 
		  Constraint::Result_t result1)
{
	Constraint::Result_t result_out;

#if 0
	g_message ("intersectResults");
#endif

	/* Take all combinations of fraglists from each one, see if they are compatible */
	for (Constraint::Result_t::iterator iter0 = result0.begin (); iter0!=result0.end(); iter0++) {
		for (Constraint::Result_t::iterator iter1 = result1.begin (); iter1!=result1.end(); iter1++) {

#if 1
			FragmentList* fraglist = combineFragmentLists (*iter0,
								       *iter1);
			if (fraglist) {
				result_out.push_back (*fraglist);
				// delete fraglist;
			}			
#endif
		}		
	}

	return result_out;
}


// class ConstraintMatch : public Constraint
Constraint::Result_t
ConstraintMatch::
generateAll () const
{
	Result_t result;

	g_assert (m_offsetStart<=m_offsetEnd);

	FragmentFixed *frag = generateMatchFrag ();
	g_assert (frag);

	for (guint32 start = m_offsetStart; start<=m_offsetEnd; start++) {
		FragmentList fraglist;

		if (start>0) {
			fraglist.push_back (new FragmentAny (start));
		}

		fraglist.push_back (frag);

		result.push_back (fraglist);
	}
	
	return result;
}

// class ConstraintMatchString : public ConstraintMatch
FragmentFixed*
ConstraintMatchString::
generateMatchFrag () const
{
	/* FIXME: encoding issues! */
	FragmentFixed *frag = new FragmentFixed();

	for (int i=0;i<m_matchValue.size();i++) {
		frag->addByte((guint8)m_matchValue[i]);
	}

	return frag;
}

// class ConstraintMatchHost16 : public ConstraintMatch
FragmentFixed*
ConstraintMatchHost16::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* FIXME: do I need to deliberately break the byte ordering?  isn't this fundamentally broken? */

	/* This should be little-endian: */
	frag->addByte((guint8)(m_matchValue>>0)&0xff);
	frag->addByte((guint8)(m_matchValue>>8)&0xff);

	return frag;
}

// class ConstraintMatchHost32 : public ConstraintMatch
FragmentFixed*
ConstraintMatchHost32::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* FIXME: do I need to deliberately break the byte ordering?  isn't this fundamentally broken? */

	/* This should be little-endian: */
	frag->addByte((guint8)(m_matchValue>>0)&0xff);
	frag->addByte((guint8)(m_matchValue>>8)&0xff);
	frag->addByte((guint8)(m_matchValue>>16)&0xff);
	frag->addByte((guint8)(m_matchValue>>24)&0xff);

	return frag;
}

// class ConstraintMatchLittle16 : public ConstraintMatch
FragmentFixed*
ConstraintMatchLittle16::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* This should be little-endian: */
	frag->addByte((guint8)(m_matchValue>>0)&0xff);
	frag->addByte((guint8)(m_matchValue>>8)&0xff);

	return frag;
}

// class ConstraintMatchLittle32 : public ConstraintMatch
FragmentFixed*
ConstraintMatchLittle32::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* This should be little-endian: */
	frag->addByte((guint8)(m_matchValue>>0)&0xff);
	frag->addByte((guint8)(m_matchValue>>8)&0xff);
	frag->addByte((guint8)(m_matchValue>>16)&0xff);
	frag->addByte((guint8)(m_matchValue>>24)&0xff);

	return frag;
}

// class ConstraintMatchBig16 : public ConstraintMatch
FragmentFixed*
ConstraintMatchBig16::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* This should be big-endian: */
	frag->addByte((guint8)(m_matchValue>>8)&0xff);
	frag->addByte((guint8)(m_matchValue>>0)&0xff);

	return frag;
}

// class ConstraintMatchBig32 : public ConstraintMatch
FragmentFixed*
ConstraintMatchBig32::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	/* This should be big-endian: */
	frag->addByte((guint8)(m_matchValue>>24)&0xff);
	frag->addByte((guint8)(m_matchValue>>16)&0xff);
	frag->addByte((guint8)(m_matchValue>>8)&0xff);
	frag->addByte((guint8)(m_matchValue>>0)&0xff);

	return frag;
}

// class ConstraintMatchByte : public ConstraintMatch
FragmentFixed*
ConstraintMatchByte::
generateMatchFrag () const
{
	FragmentFixed *frag = new FragmentFixed();

	frag->addByte(m_matchValue);

	return frag;
}

// class ConstraintAnd : public ConstraintCompound

Constraint::Result_t
ConstraintAnd::
generateAll () const
{
	Result_t result;

#if 0
	g_message("ConstraintAnd::generateAll ()");
#endif

	/* Need to start with a an empty fraglist: */
	result.push_back (FragmentList());
	
	for (ChildContainer_t::const_iterator iter = m_children.begin(); iter!=m_children.end(); iter++) {
		Result_t inner_result = (*iter)->generateAll (); 

		result = intersectResults (result, inner_result);
	}

	return result;
}

// class ConstraintOr : public ConstraintCompound
Constraint::Result_t
ConstraintOr::
generateAll () const
{
	Result_t result;
	
	for (ChildContainer_t::const_iterator iter = m_children.begin(); iter!=m_children.end(); iter++) {
		Result_t inner_result = (*iter)->generateAll (); 

		/* For now, add all results from each sub-contraint to our list; this will potentially yield duplicates: */
		result.insert (result.end(), inner_result.begin (), inner_result.end() );
	}

	return result;
}

