#include "xml-wrapper-mime.h"
#include "constraint.h"
#include "report.h"

// Predicate: can a filename be generated that will satisy the filename constaints of all the mimetypes in the input vector:
bool
can_have_compatible_filenames(std::vector<WrapMimeType> mimeTypes)
{
	// For now, assume that if more than one type in the list specifies a <glob>, then a name can't be generated that will satisfy all in the list.
	// If we were more sophisticated then we'd do some kind of pattern matching.
	
	bool found_glob = false;

	for (std::vector<WrapMimeType>::const_iterator iter = mimeTypes.begin(); iter!=mimeTypes.end(); iter++) {
		if (iter->hasChild("glob")) {
			if (found_glob) {
				// Found more than one <glob>:
				return false;
			} 

			found_glob=true;
		}		
	}

	return true;
}



// class WrapMimeDatabase : public DocWrapper
xmlNodePtr
WrapMimeDatabase::
getNodeForMime (std::string mimeType) const
{
	for (xmlNodePtr iter=m_nodeMimeInfo->children;iter;iter=iter->next) {
		if (NodeWrapper(iter).isElement ("mime-type")) {
			if (WrapMimeType(iter).asString ()==mimeType) {
				return iter;
			}
		}
	}	

	return NULL;
}


void
WrapMimeDatabase::
debug () const
{
	xmlNodePtr iter;

	g_message ("Mime Database:");

	for (iter=m_nodeMimeInfo->children;iter;iter=iter->next) {
		if (NodeWrapper(iter).isElement ("mime-type")) {
			g_message (WrapMimeType(iter).asString ().c_str());
		}
	}	
}

/* Given a non-empty vector of mime types which can clash, see if there are any ways of adding an additional mime type to the list
   that creates a further clash, and recurse.

   The input vector is assumed to be sorted in list of occurrence within the database, and it's assumed that you don't have to search for
   any mimetypes that occur before the final one in the vector

   As an optimisation, we also receive the set of all possible intersections of the input list, to prevent having to regenerate this
 */
void
WrapMimeDatabase::
findIntersectionsRecursive (std::vector<WrapMimeType> intersectingMimeTypes,
			    Constraint::Result_t inputResults,
			    Report &report) const
{
	g_assert (!intersectingMimeTypes.empty());
	g_assert (intersectingMimeTypes.size()<report.arity());

	WrapMimeType finalMimeType = intersectingMimeTypes[intersectingMimeTypes.size()-1];

	for (xmlNodePtr iter=finalMimeType.getNode()->next; iter; iter=iter->next) {
		if (NodeWrapper(iter).isElement ("mime-type")) {
			WrapMimeType mimeType (iter);

			if (report.considerMimeType(mimeType)) {
#if 1
				std::vector<WrapMimeType> newIntersectingMimeTypes = intersectingMimeTypes;
				
				newIntersectingMimeTypes.push_back (WrapMimeType(iter));
					
				
				if (report.matchNames()) {
					if (!can_have_compatible_filenames(newIntersectingMimeTypes)) {
						continue;
					}
				}

				Constraint::Result_t outputResults = intersectResults (inputResults,
										       mimeType.generateAll ());
				
				/* Report and recurse if an intersection was found: */
				if (!outputResults.empty()) {
					int arity = newIntersectingMimeTypes.size();

					/* Report: */
					if (arity==report.arity()) {
						report.onIntersections (newIntersectingMimeTypes,  
									outputResults);
					}
					
					/* Recurse: */
					if (arity<report.arity()) {
						findIntersectionsRecursive (newIntersectingMimeTypes,  
									    outputResults,
									    report);
					}
				}
			}
#else
			std::vector<WrapMimeType> mimeTypes = intersectingMimeTypes;
			
			mimeTypes.push_back (WrapMimeType(iter));
			
			Constraint::Result_t outputResults = WrapMimeType::findIntersection (mimeTypes,
											 report);
			if (!outputResults.empty()) {
				/* Recurse if an intersection was found: */
				findIntersectionsRecursive (mimeTypes,
							    outputResults,
							    report);
			}
#endif
		}

	}	
}

void
WrapMimeDatabase::
findIntersections (Report& report) const
{
	xmlNodePtr iter1, iter2;

	for (iter1=m_nodeMimeInfo->children;iter1;iter1=iter1->next) {
		if (NodeWrapper(iter1).isElement ("mime-type")) {

			WrapMimeType mime1(iter1);

			if (report.considerMimeType(mime1)) {
				g_message ("Finding intersections with %s ", mime1.asString ().c_str());
				
				for (iter2=iter1->next;iter2;iter2=iter2->next) {
					if (NodeWrapper(iter2).isElement ("mime-type")) {
						WrapMimeType mime2(iter2);

						if (report.considerMimeType(mime2)) {
							std::vector<WrapMimeType> mimeTypes;
							
							mimeTypes.push_back (mime1);
							mimeTypes.push_back (mime2);
							
							if (report.matchNames()) {
								if (!can_have_compatible_filenames(mimeTypes)) {
									continue;
								}
							}
							
							Constraint::Result_t outputResults = WrapMimeType::findIntersection (mimeTypes,
															     report);
							if (!outputResults.empty()) {
								/* Report: */
								if (report.arity()==2) {
									/* Report: */
									report.onIntersections (mimeTypes,  
												outputResults);
								}
								
								if (report.arity()>2) {
									/* Start recursion to try to find 3-way intersections (and beyond): */
									findIntersectionsRecursive (mimeTypes,
												    outputResults,
												    report);
								}
							}
						}
					}
				}
			}
		}
	}
}



// class WrapMimeType : public NodeWrapper

std::string
WrapMimeType::
asString () const
{
	return getAttrValue ("type");
}

bool
WrapMimeType::
hasMagic () const
{
	return NULL!=locateChildElement ("magic");
}

bool
WrapMimeType::
hasConstraints () const
{
	return hasMagic()||hasChild("glob");
}

void
dumpResults (Constraint::Result_t results);

Constraint*
WrapMimeType::
makeConstraint () const
{
	xmlNodePtr nodeMagic = locateChildElement ("magic");
		
	/* FIXME: support multiple magic elements per WrapMimeType */
	
	if (nodeMagic) {			
		return WrapMimeMagic (nodeMagic).makeConstraint ();
	} else {
		return NULL;
	}
}

Constraint::Result_t
WrapMimeType::
findIntersection (std::vector<WrapMimeType> mimeTypes,
		  Report& report)
{
	ConstraintAnd *intersection = new ConstraintAnd;
	
	for (std::vector<WrapMimeType>::iterator iter = mimeTypes.begin(); iter!=mimeTypes.end(); iter++) {
		Constraint *constraint = iter->makeConstraint ();

		if (constraint) {			
			intersection->addChild (constraint);
		}
	}
	
	intersection->optimise ();
	
	Constraint::Result_t results = intersection->generateAll ();
	
	if (!results.empty()) {
		report.onIntersections (mimeTypes,  
					results);
		
	}

	return results;

}

Constraint::Result_t
WrapMimeType::
generateAll () const
{
	RefCountPtr<Constraint> constraint = makeConstraint ();

	g_assert (constraint.isNonNull ());

	return constraint->generateAll ();
}

#if 0
/* Given an intersecting collection of mimetypes and the set of fragment lists that are their intersection, 
   see if they intersect with this mimetype, and generate the resulting fragment list */
Constraint::Result_t
WrapMimeType::
findIntersectionOptimised (std::vector<WrapMimeType> mimeTypes,
			   Constraint::Result_t inputResults) const
{
#error
	return intersectResults (inputResults,
				 generateAll ());
	
}
#endif

// class WrapMimeMagic : public NodeWrapper
Constraint*
WrapMimeMagic::
makeConstraint () const
{
	ConstraintOr *constraint = new ConstraintOr ();

	xmlNodePtr iter;

	for (iter=m_xmlNode->children;iter;iter=iter->next) {
		if (NodeWrapper(iter).isElement ("match")) {
			
       			constraint->addChild (WrapMimeMatch(iter).makeConstraint ());
		}
	}

	return constraint;
}

int
WrapMimeMagic::
getPriority () const
{
	/* use default value of 0 */
	return getAttrAsInt ("priority",
			     0);
}
		
// class WrapMimeMatch : public NodeWrapper
Constraint*
WrapMimeMatch::
makeConstraint () const
{
	/* FIXME: what about nested <match> elements? are they treated as "and"? */
	std::string type = getAttrValue ("type");
	std::string offset = getAttrValue ("offset");
	std::string value = getAttrValue ("value");

	/* FIXME: we don't support mask */
	
	guint offsetStart;
	guint offsetEnd;

	guint numericValue = 0;

	calcOffsets (offset, offsetStart, offsetEnd);

	sscanf (value.c_str(), "%i", &numericValue);

	if (type=="string") {
		return new ConstraintMatchString (offsetStart,
						  offsetEnd,
						  value);
	} else if (type=="host16") {
		return new ConstraintMatchHost16 (offsetStart,
						  offsetEnd,
						  numericValue);
	} else if (type=="host32") {
		return new ConstraintMatchHost32 (offsetStart,
						  offsetEnd,
						  numericValue);
	} else if (type=="big16") {
		return new ConstraintMatchBig16 (offsetStart,
						  offsetEnd,
						  numericValue);
	} else if (type=="big32") {
		return new ConstraintMatchBig32 (offsetStart,
						 offsetEnd,
						 numericValue);
	} else if (type=="little16") {
		return new ConstraintMatchLittle16 (offsetStart,
						    offsetEnd,
						    numericValue);
	} else if (type=="little32") {
		return new ConstraintMatchLittle32 (offsetStart,
						    offsetEnd,
						    numericValue);
	} else if (type=="byte") {
		return new ConstraintMatchByte (offsetStart,
						offsetEnd,
						numericValue);
	} else {

		g_assert_not_reached();
	}
}

void
WrapMimeMatch::
calcOffsets (std::string attrOffset,
	     guint &offsetStart,
	     guint &offsetEnd)
{
	int tmpS, tmpE;

	offsetStart = 0;
	offsetEnd = 0;

	if (2==sscanf (attrOffset.c_str(), "%i:%i", &tmpS, &tmpE)) {
		offsetStart = tmpS;
		offsetEnd = tmpE;
	} else {
		sscanf (attrOffset.c_str(), "%i", &tmpS);
		offsetStart = tmpS;
		offsetEnd = tmpS;
	}
}

